# -*- coding: utf-8 -*-
"""A class to represent a users' keys and functions to manage them."""
from __future__ import (unicode_literals, print_function)

import os
import shlex

from creds.constants import RANDOM_FILE_EXT_LENGTH
from creds.utils import base64decode, base64encode
from creds.utils import execute_command, random_string, sudo_check
from external.six import text_type


class PublicKey(object):

    """Representation of a public key."""

    def __init__(self, raw=None, b64encoded=None):
        """Make a public key.

        args:
            raw (str): raw public key
            b64encoded (str): base64 encoded public key
        """
        if not any((raw, b64encoded)):
            raise AttributeError('Key not provided')
        self._raw = raw
        self._b64encoded = b64encoded

    @property
    def b64encoded(self):
        """Return a base64 encoding of the key.

        returns:
            str: base64 encoding of the public key
        """
        if self._b64encoded:
            return text_type(self._b64encoded).strip("\r\n")
        else:
            return base64encode(self.raw)

    @property
    def raw(self):
        """Return raw key.

        returns:
            str: raw key
        """
        if self._raw:
            return text_type(self._raw).strip("\r\n")
        else:
            return text_type(base64decode(self._b64encoded)).strip("\r\n")

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return 'PublicKey(raw=\"{0}\", b64encoded=\"{1}\"'.format(self.raw, self.b64encoded)


# TODO: Keep temporary copy so we can check for race condition.

def read_authorized_keys(username=None):
    """Read public keys from specified user's authorized_keys file.

    args:
        username (str): username.

    returns:
        list: Authorised keys for the specified user.
    """
    authorized_keys_path = '{0}/.ssh/authorized_keys'.format(os.path.expanduser('~{0}'.format(username)))
    rnd_chars = random_string(length=RANDOM_FILE_EXT_LENGTH)
    tmp_authorized_keys_path = '/tmp/authorized_keys_{0}_{1}'.format(username, rnd_chars)
    authorized_keys = list()
    copy_result = execute_command(
        shlex.split(str('{0} cp {1} {2}'.format(sudo_check(), authorized_keys_path, tmp_authorized_keys_path))))
    result_message = copy_result[0][1].decode('UTF-8')
    if 'you must have a tty to run sudo' in result_message:  # pragma: no cover
        raise OSError("/etc/sudoers is blocked sudo. Remove entry: 'Defaults    requiretty'.")
    elif 'No such file or directory' not in result_message:
        execute_command(shlex.split(str('{0} chmod 755 {1}'.format(sudo_check(), tmp_authorized_keys_path))))
        with open(tmp_authorized_keys_path) as keys_file:
            for key in keys_file:
                authorized_keys.append(PublicKey(raw=key))
        execute_command(shlex.split(str('{0} rm {1}'.format(sudo_check(), tmp_authorized_keys_path))))
    return authorized_keys


def write_authorized_keys(user=None):
    """Write public keys back to authorized_keys file. Create keys directory if it doesn't already exist.

    args:
        user (User): Instance of User containing keys.

    returns:
        list: Authorised keys for the specified user.
    """
    authorized_keys = list()
    authorized_keys_dir = '{0}/.ssh'.format(os.path.expanduser('~{0}'.format(user.name)))
    rnd_chars = random_string(length=RANDOM_FILE_EXT_LENGTH)
    authorized_keys_path = '{0}/authorized_keys'.format(authorized_keys_dir)
    tmp_authorized_keys_path = '/tmp/authorized_keys_{0}_{1}'.format(user.name, rnd_chars)

    if not os.path.isdir(authorized_keys_dir):
        execute_command(shlex.split(str('{0} mkdir -p {1}'.format(sudo_check(), authorized_keys_dir))))
    for key in user.public_keys:
        authorized_keys.append('{0}\n'.format(key.raw))
    with open(tmp_authorized_keys_path, mode=text_type('w+')) as keys_file:
        keys_file.writelines(authorized_keys)
    execute_command(
        shlex.split(str('{0} cp {1} {2}'.format(sudo_check(), tmp_authorized_keys_path, authorized_keys_path))))
    execute_command(shlex.split(str('{0} chown -R {1} {2}'.format(sudo_check(), user.name, authorized_keys_dir))))
    execute_command(shlex.split(str('{0} chmod 700 {1}'.format(sudo_check(), authorized_keys_dir))))
    execute_command(shlex.split(str('{0} chmod 600 {1}'.format(sudo_check(), authorized_keys_path))))
    execute_command(shlex.split(str('{0} rm {1}'.format(sudo_check(), tmp_authorized_keys_path))))




# -*- coding: utf-8 -*-

from typing import Tuple, List, Optional, Union


def sudo_check() -> str: pass


def get_platform() -> str: pass


def get_missing_commands(_platform: str) -> List: pass


def execute_command(command: List) -> Tuple: pass


def random_string(length: Optional[int]) -> str: pass


def base64encode(_input: Union[bytes, str]) -> str: pass


def base64decode(_input: Union[bytes, str]) -> str: pass


def remove_sudoers_entry(username=Optional[str]) -> None: pass


def write_sudoers_entry(username=Optional[str], sudoers_entry=Optional[str]) -> None: pass


def read_sudoers() -> List: pass


def get_sudoers_entry(username=Optional[str], sudoers_entries=List) -> str: pass



# -*- coding: utf-8 -*-
"""This module contains common helper functions."""

from __future__ import unicode_literals

import base64
import os
import platform
import random
import shlex
import string
import subprocess

from creds.constants import (CMD_SUDO, RANDOM_FILE_EXT_LENGTH, LINUX_CMD_GROUP_ADD, LINUX_CMD_GROUP_DEL,
                             LINUX_CMD_USERADD, LINUX_CMD_USERDEL, LINUX_CMD_USERMOD, FREEBSD_CMD_PW, LINUX_CMD_VISUDO)
from external.six import (PY2, PY3, text_type)


def sudo_check():
    """Return the string 'sudo' if current user isn't root."""
    sudo_cmd = ''
    if os.geteuid() != 0:
        sudo_cmd = CMD_SUDO
    return sudo_cmd


def get_platform():
    """Return platform name"""
    return platform.system()


def get_missing_commands(_platform):
    """Check I can identify the necessary commands for managing users."""
    missing = list()
    if _platform in ('Linux', 'OpenBSD'):
        if not LINUX_CMD_USERADD:
            missing.append('useradd')
        if not LINUX_CMD_USERMOD:
            missing.append('usermod')
        if not LINUX_CMD_USERDEL:
            missing.append('userdel')
        if not LINUX_CMD_GROUP_ADD:
            missing.append('groupadd')
        if not LINUX_CMD_GROUP_DEL:
            missing.append('groupdel')
    elif _platform == 'FreeBSD':  # pragma: FreeBSD
        # FREEBSD COMMANDS
        if not FREEBSD_CMD_PW:
            missing.append('pw')
    if missing:
        print('\nMISSING = {0}'.format(missing))
    return missing


def execute_command(command=None):
    """Execute a command and return the stdout and stderr."""
    process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    stdout, stdin = process.communicate()
    process.wait()
    return (stdout, stdin), process.returncode


def random_string(length=None):
    """Generate a random string of ASCII characters."""
    return ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits)
                   for _ in range(length))


def base64encode(_input=None):
    """Return base64 encoded representation of a string."""
    if PY2:  # pragma: no cover
        return base64.b64encode(_input)
    elif PY3:  # pragma: no cover
        if isinstance(_input, bytes):
            return base64.b64encode(_input).decode('UTF-8')
        elif isinstance(_input, str):
            return base64.b64encode(bytearray(_input, encoding='UTF-8')).decode('UTF-8')


def base64decode(_input=None):
    """Take a base64 encoded string and return the decoded string."""
    missing_padding = 4 - len(_input) % 4
    if missing_padding:
        _input += '=' * missing_padding
    if PY2:  # pragma: no cover
        return base64.decodestring(_input)
    elif PY3:  # pragma: no cover
        if isinstance(_input, bytes):
            return base64.b64decode(_input).decode('UTF-8')
        elif isinstance(_input, str):
            return base64.b64decode(bytearray(_input, encoding='UTF-8')).decode('UTF-8')


def read_sudoers():
    """ Read the sudoers entry for the specified user.

    args:
        username (str): username.

    returns:`r
        str: sudoers entry for the specified user.
    """
    sudoers_path = '/etc/sudoers'
    rnd_chars = random_string(length=RANDOM_FILE_EXT_LENGTH)
    tmp_sudoers_path = '/tmp/sudoers_{0}'.format(rnd_chars)
    sudoers_entries = list()
    copy_result = execute_command(
        shlex.split(str('{0} cp {1} {2}'.format(sudo_check(), sudoers_path, tmp_sudoers_path))))
    result_message = copy_result[0][1].decode('UTF-8')
    if 'No such file or directory' not in result_message:
        execute_command(shlex.split(str('{0} chmod 755 {1}'.format(sudo_check(), tmp_sudoers_path))))
        with open(tmp_sudoers_path) as tmp_sudoers_file:
            for line in tmp_sudoers_file:
                stripped = line.strip().replace(os.linesep, '')
                if stripped and not stripped.startswith('#'):
                    sudoers_entries.append(stripped)
        execute_command(shlex.split(str('{0} rm {1}'.format(sudo_check(), tmp_sudoers_path))))
    return sudoers_entries


def write_sudoers_entry(username=None, sudoers_entry=None):
    """Write sudoers entry.

    args:
        user (User): Instance of User containing sudoers entry.

    returns:
        str: sudoers entry for the specified user.
    """

    sudoers_path = '/etc/sudoers'
    rnd_chars = random_string(length=RANDOM_FILE_EXT_LENGTH)
    tmp_sudoers_path = '/tmp/sudoers_{0}'.format(rnd_chars)
    execute_command(
        shlex.split(str('{0} cp {1} {2}'.format(sudo_check(), sudoers_path, tmp_sudoers_path))))
    execute_command(
        shlex.split(str('{0} chmod 777 {1}'.format(sudo_check(), tmp_sudoers_path))))
    with open(tmp_sudoers_path, mode=text_type('r')) as tmp_sudoers_file:
        sudoers_entries = tmp_sudoers_file.readlines()
    sudoers_output = list()
    for entry in sudoers_entries:
        if entry and not entry.startswith(username):
            sudoers_output.append(entry)
    if sudoers_entry:
        sudoers_output.append('{0} {1}'.format(username, sudoers_entry))
        sudoers_output.append('\n')
    with open(tmp_sudoers_path, mode=text_type('w+')) as tmp_sudoers_file:
        tmp_sudoers_file.writelines(sudoers_output)
    sudoers_check_result = execute_command(
        shlex.split(str('{0} {1} -cf {2}'.format(sudo_check(), LINUX_CMD_VISUDO, tmp_sudoers_path))))
    if sudoers_check_result[1] > 0:
        raise ValueError(sudoers_check_result[0][1])
    execute_command(
        shlex.split(str('{0} cp {1} {2}'.format(sudo_check(), tmp_sudoers_path, sudoers_path))))
    execute_command(shlex.split(str('{0} chown root:root {1}'.format(sudo_check(), sudoers_path))))
    execute_command(shlex.split(str('{0} chmod 440 {1}'.format(sudo_check(), sudoers_path))))
    execute_command(shlex.split(str('{0} rm {1}'.format(sudo_check(), tmp_sudoers_path))))


def remove_sudoers_entry(username=None):
    """Remove sudoers entry.

    args:
        user (User): Instance of User containing sudoers entry.

    returns:
        str: sudoers entry for the specified user.
    """
    sudoers_path = '/etc/sudoers'
    rnd_chars = random_string(length=RANDOM_FILE_EXT_LENGTH)
    tmp_sudoers_path = '/tmp/sudoers_{0}'.format(rnd_chars)
    execute_command(
        shlex.split(str('{0} cp {1} {2}'.format(sudo_check(), sudoers_path, tmp_sudoers_path))))
    execute_command(
        shlex.split(str('{0} chmod 777 {1}'.format(sudo_check(), tmp_sudoers_path))))
    with open(tmp_sudoers_path, mode=text_type('r')) as tmp_sudoers_file:
        sudoers_entries = tmp_sudoers_file.readlines()
    sudoers_output = list()
    for entry in sudoers_entries:
        if not entry.startswith(username):
            sudoers_output.append(entry)
    with open(tmp_sudoers_path, mode=text_type('w+')) as tmp_sudoers_file:
        tmp_sudoers_file.writelines(sudoers_output)
    execute_command(
        shlex.split(str('{0} cp {1} {2}'.format(sudo_check(), tmp_sudoers_path, sudoers_path))))
    execute_command(shlex.split(str('{0} chown root:root {1}'.format(sudo_check(), sudoers_path))))
    execute_command(shlex.split(str('{0} chmod 440 {1}'.format(sudo_check(), sudoers_path))))
    execute_command(shlex.split(str('{0} rm {1}'.format(sudo_check(), tmp_sudoers_path))))


def get_sudoers_entry(username=None, sudoers_entries=None):
    """ Find the sudoers entry in the sudoers file for the specified user.

    args:
        username (str): username.
        sudoers_entries (list): list of lines from the sudoers file.

    returns:`r
        str: sudoers entry for the specified user.
    """
    for entry in sudoers_entries:
        if entry.startswith(username):
            return entry.replace(username, '').strip()



# -*- coding: utf-8 -*-
"""Functions to generate a list of steps to transition from the current state to the desired state."""
from __future__ import (unicode_literals, print_function)

from creds import constants
from creds.ssh import write_authorized_keys
from creds.users import (generate_add_user_command, generate_modify_user_command,
                         generate_delete_user_command, compare_user, get_user_by_uid)
from creds.utils import execute_command, write_sudoers_entry, remove_sudoers_entry
from external.six import iteritems


def create_plan(existing_users=None, proposed_users=None, purge_undefined=None, protected_users=None,
                allow_non_unique_id=None, manage_home=True, manage_keys=True):
    """Determine what changes are required.

    args:
        existing_users (Users): List of discovered users
        proposed_users (Users): List of proposed users
        purge_undefined (bool): Remove discovered users that have not been defined in proposed users list
        protected_users (list): List of users' names that should not be evaluated as part of the plan creation process
        allow_non_unique_id (bool): Allow more than one user to have the same uid
        manage_home (bool): Create/remove users' home directories
        manage_keys (bool): Add/update/remove users' keys (manage_home must also be true)

    returns:
       list: Differences between discovered and proposed users with a
             list of operations that will achieve the desired state.
    """

    plan = list()
    proposed_usernames = list()

    if not purge_undefined:
        purge_undefined = constants.PURGE_UNDEFINED
    if not protected_users:
        protected_users = constants.PROTECTED_USERS
    if not allow_non_unique_id:
        allow_non_unique_id = constants.ALLOW_NON_UNIQUE_ID

    # Create list of modifications to make based on proposed users compared to existing users
    for proposed_user in proposed_users:
        proposed_usernames.append(proposed_user.name)
        user_matching_name = existing_users.describe_users(users_filter=dict(name=proposed_user.name))
        user_matching_id = get_user_by_uid(uid=proposed_user.uid, users=existing_users)
        # If user does not exist
        if not allow_non_unique_id and user_matching_id and not user_matching_name:
            plan.append(
                dict(action='fail', error='uid_clash', proposed_user=proposed_user, state='existing', result=None))
        elif not user_matching_name:
            plan.append(
                dict(action='add', proposed_user=proposed_user, state='missing', result=None, manage_home=manage_home,
                     manage_keys=manage_keys))
        # If they do, then compare
        else:
            user_comparison = compare_user(passed_user=proposed_user, user_list=existing_users)
            if user_comparison.get('result'):
                plan.append(
                    dict(action='update', proposed_user=proposed_user, state='existing',
                         user_comparison=user_comparison, manage_home=manage_home, manage_keys=manage_keys))
    # Application of the proposed user list will not result in deletion of users that need to be removed
    # If 'PURGE_UNDEFINED' then look for existing users that are not defined in proposed usernames and mark for removal
    if purge_undefined:
        for existing_user in existing_users:
            if existing_user.name not in proposed_usernames:
                if existing_user.name not in protected_users:
                    plan.append(
                        dict(action='delete', username=existing_user.name, state='existing', manage_home=manage_home,
                             manage_keys=manage_keys))
    return plan


def execute_plan(plan=None):
    """Create, Modify or Delete, depending on plan item."""
    execution_result = list()
    for task in plan:
        action = task['action']
        if action == 'delete':
            command = generate_delete_user_command(username=task.get('username'), manage_home=task['manage_home'])
            command_output = execute_command(command)
            execution_result.append(dict(task=task, command_output=command_output))
            remove_sudoers_entry(username=task.get('username'))
        elif action == 'add':
            command = generate_add_user_command(proposed_user=task.get('proposed_user'), manage_home=task['manage_home'])
            command_output = execute_command(command)
            if task['proposed_user'].public_keys and task['manage_home'] and task['manage_keys']:
                write_authorized_keys(task['proposed_user'])
            if task['proposed_user'].sudoers_entry:
                write_sudoers_entry(username=task['proposed_user'].name,
                                    sudoers_entry=task['proposed_user'].sudoers_entry)
            execution_result.append(dict(task=task, command_output=command_output))
        elif action == 'update':
            result = task['user_comparison'].get('result')
            # Don't modify user if only keys have changed
            action_count = 0
            for k, _ in iteritems(result):
                if '_action' in k:
                    action_count += 1
            command_output = None
            if task['manage_home'] and task['manage_keys'] and action_count == 1 and 'public_keys_action' in result:
                write_authorized_keys(task['proposed_user'])
            elif action_count == 1 and 'sudoers_entry_action' in result:
                write_sudoers_entry(username=task['proposed_user'].name,
                                    sudoers_entry=task['user_comparison']['result']['replacement_sudoers_entry'])
            else:
                command = generate_modify_user_command(task=task)
                command_output = execute_command(command)
                if task['manage_home'] and task['manage_keys'] and result.get('public_keys_action'):
                    write_authorized_keys(task['proposed_user'])
                if result.get('sudoers_entry_action'):
                    write_sudoers_entry(username=task['proposed_user'].name,
                                        sudoers_entry=task['user_comparison']['result']['replacement_sudoers_entry'])
            execution_result.append(dict(task=task, command_output=command_output))



# -*- coding: utf-8 -*-
"""Functions to define and discover OS constants."""
from __future__ import (unicode_literals, print_function)

import io
import os
from distutils import spawn

from external.six import (text_type, PY2, PY3)

SUPPORTED_PLATFORMS = ['Linux', 'FreeBSD', 'OpenBSD']

RANDOM_FILE_EXT_LENGTH = 20  # Number of random characters to add to transient file names

PURGE_UNDEFINED = False  # Purge any users that fall between UID_MIN and UID_MAX that are not defined

DEFAULT_UID_MIN = 1000  # The lowest uid to consider safe to manage
DEFAULT_UID_MAX = 60000  # The maximum uid to consider safe to manage

CMD_SUDO = spawn.find_executable("sudo")

# LINUX/OPENBSD COMMANDS
LINUX_CMD_USERADD = spawn.find_executable("useradd")
LINUX_CMD_USERMOD = spawn.find_executable("usermod")
LINUX_CMD_USERDEL = spawn.find_executable("userdel")
LINUX_CMD_GROUP_ADD = spawn.find_executable("groupadd")
LINUX_CMD_GROUP_DEL = spawn.find_executable("groupdel")
LINUX_CMD_VISUDO = spawn.find_executable("visudo")
# FREEBSD COMMANDS
FREEBSD_CMD_PW = spawn.find_executable("pw")


def login_defs():
    """Discover the minimum and maximum UID number."""
    uid_min = None
    uid_max = None
    login_defs_path = '/etc/login.defs'
    if os.path.exists(login_defs_path):
        with io.open(text_type(login_defs_path), encoding=text_type('utf-8')) as log_defs_file:
            login_data = log_defs_file.readlines()
        for line in login_data:
            if PY3:  # pragma: no cover
                line = str(line)
            if PY2:  # pragma: no cover
                line = line.encode(text_type('utf8'))
            if line[:7] == text_type('UID_MIN'):
                uid_min = int(line.split()[1].strip())
            if line[:7] == text_type('UID_MAX'):
                uid_max = int(line.split()[1].strip())
    if not uid_min:  # pragma: no cover
        uid_min = DEFAULT_UID_MIN
    if not uid_max:  # pragma: no cover
        uid_max = DEFAULT_UID_MAX
    return uid_min, uid_max


UID_MIN, UID_MAX = login_defs()
ALLOW_NON_UNIQUE_ID = False  # Allow multiple users to share uids
PROTECTED_USERS = list()  # Users that must not be affected



# -*- coding: utf-8 -*-
from __future__ import (unicode_literals, print_function)

from typing import AnyStr, Optional, List

from creds.users import User


class PublicKey(object):
    def __init__(self, raw: Optional[AnyStr], b64encoded: Optional[AnyStr]) -> None:
        self._raw = raw
        self._b64encoded = b64encoded

    @property
    def b64encoded(self) -> Optional[AnyStr]: pass

    @property
    def raw(self) -> Optional[AnyStr]: pass


# TODO: Keep temporary copy so we can check for race condition.

def read_authorized_keys(username: str) -> List: pass


def write_authorized_keys(user: User) -> List: pass



# -*- coding: utf-8 -*-
from typing import List, Dict, AnyStr, Optional, Any, MutableSequence

from creds.ssh import PublicKey


class User(object):
    def __init__(self, name: AnyStr, passwd: Optional[AnyStr], uid: Optional[int], gid: Optional[int],
                 gecos: Optional[AnyStr], home_dir: Optional[AnyStr],
                 shell: Optional[AnyStr], public_keys: Optional[List[PublicKey]],
                 sudoers_entry: Optional[AnyStr]) -> None:
        self.name = name
        self.passwd = passwd
        self.uid = uid
        self.gid = gid
        self._gecos = gecos
        self.home_dir = home_dir
        self.shell = shell
        self.public_keys = public_keys
        self.sudoers_entry = sudoers_entry

    def gecos(self) -> str: pass

    def to_dict(self) -> dict: pass


class Users(MutableSequence):
    def __init__(self) -> None:
        self._user_list = list()
        self.oktypes = User

    def describe_users(self, users_filter: Dict) -> List: pass

    def check(self, value: Any) -> None: pass

    def insert(self, index: int, value: User) -> None: pass

    def remove(self, username: str) -> None: pass

    def __iter__(self) -> User: pass

    def __len__(self) -> int: pass

    def __getitem__(self, index: int) -> User: pass

    def __setitem__(self, index: int, value: User) -> None: pass

    def __delitem__(self, index: int) -> None: pass

    @classmethod
    def from_dict(cls, input_dict: Dict) -> List: pass

    @classmethod
    def from_yaml(cls, file_path: str) -> List: pass

    @classmethod
    def from_json(cls, file_path: str) -> List: pass

    @staticmethod
    def from_passwd(uid_min: int, uid_max: int) -> Users: pass

    @staticmethod
    def construct_user_list(raw_users: dict) -> Users: pass


def generate_add_user_command(proposed_user: User, manage_home: bool) -> List[str]: pass


def generate_modify_user_command(task: dict) -> List[str]: pass


def compare_user(passed_user: User, user_list=Users) -> Dict: pass


def get_user_by_uid(uid: int, users: Users) -> Users: pass



# -*- coding: utf-8 -*-
"""
This package contains all of the modules utilised by the creds library.

constants: Functions to define and discover OS constants.

plan: Functions to generate a list of steps to transition from the current state to the desired state.

ssh: Contains a class to represent a users' keys and functions to manage them.

users: Functions  to managing users and classes...
 Users - representation of a user list and methods to read, write and manage them
 User - representation of a single user with their associated credentials

utils: Common helper functions.
"""

from __future__ import (unicode_literals, print_function)

__title__ = 'creds'
__version__ = '1.0.16'
__author__ = 'Jon Hadfield'
__license__ = 'MIT'
__copyright__ = 'Copyright 2016 Jon Hadfield'



# -*- coding: utf-8 -*-
from typing import List

from creds.users import Users


def create_plan(existing_users: Users, proposed_users: Users, purge_undefined: bool,
                protected_users: List[str],
                allow_non_unique_id: bool, manage_home: bool, manage_keys: bool) -> List: pass


def execute_plan(plan: List[dict]) -> None: pass



# -*- coding: utf-8 -*-
"""This module contains the classes for User (a user's details) and Users (a collection of User instances)."""
from __future__ import unicode_literals

import io
import json
import os
import shlex
import sys

if sys.version_info[:2] >= (3, 8):
    from collections.abc import MutableSequence
else:
    from collections import MutableSequence

from creds.constants import (SUPPORTED_PLATFORMS, UID_MAX, UID_MIN,
                             LINUX_CMD_USERADD, LINUX_CMD_USERDEL, LINUX_CMD_USERMOD,
                             FREEBSD_CMD_PW)
from creds.ssh import PublicKey
from creds.ssh import read_authorized_keys
from creds.utils import (get_platform, sudo_check, read_sudoers, get_sudoers_entry, get_missing_commands)
from external.six import text_type


class User(object):
    """Representation of a user and their related credentials."""

    def __init__(self, name=None, passwd=None, uid=None, gid=None, gecos=None,
                 home_dir=None, shell=None, public_keys=None, sudoers_entry=None):
        """Make a user.

        args:
            name (str): user name.
            passwd (str, optional): password
            uid (int, optional): user id
            gid (int, optional): group id
            gecos (str): GECOS field
            home_dir (str): home directory
            shell (str): shell
            public_keys (list): list of public key instances
            sudoers_entry (str): an entry in sudoers
        """
        self.name = name
        self.passwd = passwd
        self.uid = uid
        self.gid = gid
        self._gecos = gecos
        self.home_dir = home_dir
        self.shell = shell
        self.public_keys = public_keys
        self.sudoers_entry = sudoers_entry

    @property
    def gecos(self):
        """Force double quoted gecos.

        returns:
            str: The double quoted gecos.
        """
        if not self._gecos:
            return None
        if self._gecos.startswith(text_type('\'')) and self._gecos.endswith(text_type('\'')):
            self._gecos = '\"{0}\"'.format(self._gecos[1:-1])
            return self._gecos
        elif self._gecos.startswith(text_type('\"')) and self._gecos.endswith(text_type('\"')):
            return self._gecos
        else:
            return '\"{0}\"'.format(self._gecos)

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<User {0}>'.format(self.name)

    def to_dict(self):
        """ Return the user as a dict. """
        public_keys = [public_key.b64encoded for public_key in self.public_keys]
        return dict(name=self.name, passwd=self.passwd, uid=self.uid, gid=self.gid, gecos=self.gecos,
                    home_dir=self.home_dir, shell=self.shell, public_keys=public_keys)


class Users(MutableSequence):
    """A collection of users and methods to manage them."""

    def __init__(self, oktypes=User):
        """Create instance of Users collection.

        args:
            oktypes (type): The acceptable types of instances..
        """
        platform = get_platform()
        # Check platform is supported
        if not platform in SUPPORTED_PLATFORMS:
            sys.exit('Linux, FreeBSD and OpenBSD are currently the only supported platforms for this library.')
        # Check OS commands are available for managing users
        missing_commands = get_missing_commands(platform)
        if missing_commands:
            sys.exit('Unable to find commands: {0}.\nPlease check PATH.'.format(', '.join(missing_commands)))

        self.oktypes = oktypes
        self._user_list = list()

    def check(self, value):
        """Check types."""
        if not isinstance(value, self.oktypes):
            raise TypeError

    def __iter__(self):
        for user in self._user_list:
            yield user

    def __len__(self):
        return len(self._user_list)

    def __str__(self):
        return self.__repr__()

    def __getitem__(self, index):
        return self._user_list[index]

    def insert(self, index, value):
        """Insert an instance of User into the collection."""
        self.check(value)
        self._user_list.insert(index, value)

    def __setitem__(self, index, value):
        self.check(value)
        self._user_list[index] = value

    def __repr__(self):
        user_list = ['{0}'.format(user) for user in self._user_list]
        output = '\n'.join(user_list)
        return output

    def __delitem__(self, index):
        del self._user_list[index]

    def remove(self, username=None):
        """Remove User instance based on supplied user name."""
        self._user_list = [user for user in self._user_list if user.name != username]

    def describe_users(self, users_filter=None):
        """Return a list of users matching a filter (if provided)."""
        user_list = Users(oktypes=User)
        for user in self._user_list:
            if users_filter and (users_filter.get('name') == user.name or users_filter.get('uid') == user.uid):
                user_list.append(user)
        return user_list

    @classmethod
    def from_dict(cls, input_dict=None):
        """Create collection from dictionary content."""
        return cls.construct_user_list(raw_users=input_dict.get('users'))

    @classmethod
    def from_yaml(cls, file_path=None):
        """Create collection from a YAML file."""
        try:
            import yaml
        except ImportError:  # pragma: no cover
            yaml = None
        if not yaml:
            import sys
            sys.exit('PyYAML is not installed, but is required in order to parse YAML files.'
                     '\nTo install, run:\n$ pip install PyYAML\nor visit'
                     ' http://pyyaml.org/wiki/PyYAML for instructions.')

        with io.open(file_path, encoding=text_type('utf-8')) as stream:
            users_yaml = yaml.safe_load(stream)
            if isinstance(users_yaml, dict):
                return cls.construct_user_list(raw_users=users_yaml.get('users'))
            else:
                raise ValueError('No YAML object could be decoded')

    @classmethod
    def from_json(cls, file_path=None):
        """Create collection from a JSON file."""
        with io.open(file_path, encoding=text_type('utf-8')) as stream:
            try:
                users_json = json.load(stream)
            except ValueError:
                raise ValueError('No JSON object could be decoded')
            return cls.construct_user_list(raw_users=users_json.get('users'))

    @staticmethod
    def from_passwd(uid_min=None, uid_max=None):
        """Create collection from locally discovered data, e.g. /etc/passwd."""
        import pwd
        users = Users(oktypes=User)
        passwd_list = pwd.getpwall()
        if not uid_min:
            uid_min = UID_MIN
        if not uid_max:
            uid_max = UID_MAX
        sudoers_entries = read_sudoers()
        for pwd_entry in passwd_list:
            if uid_min <= pwd_entry.pw_uid <= uid_max:
                user = User(name=text_type(pwd_entry.pw_name),
                            passwd=text_type(pwd_entry.pw_passwd),
                            uid=pwd_entry.pw_uid,
                            gid=pwd_entry.pw_gid,
                            gecos=text_type(pwd_entry.pw_gecos),
                            home_dir=text_type(pwd_entry.pw_dir),
                            shell=text_type(pwd_entry.pw_shell),
                            public_keys=read_authorized_keys(username=pwd_entry.pw_name),
                            sudoers_entry=get_sudoers_entry(username=pwd_entry.pw_name,
                                                            sudoers_entries=sudoers_entries))
                users.append(user)
        return users

    @staticmethod
    def construct_user_list(raw_users=None):
        """Construct a list of User objects from a list of dicts."""
        users = Users(oktypes=User)
        for user_dict in raw_users:
            public_keys = None
            if user_dict.get('public_keys'):
                public_keys = [PublicKey(b64encoded=x, raw=None)
                               for x in user_dict.get('public_keys')]
            users.append(User(name=user_dict.get('name'),
                              passwd=user_dict.get('passwd'),
                              uid=user_dict.get('uid'),
                              gid=user_dict.get('gid'),
                              home_dir=user_dict.get('home_dir'),
                              gecos=user_dict.get('gecos'),
                              shell=user_dict.get('shell'),
                              public_keys=public_keys,
                              sudoers_entry=user_dict.get('sudoers_entry')))
        return users

    def to_dict(self):
        """ Return a dict of the users. """
        users = dict(users=list())
        for user in self:
            users['users'].append(user.to_dict())
        return users

    def export(self, file_path=None, export_format=None):
        """ Write the users to a file. """
        with io.open(file_path, mode='w', encoding="utf-8") as export_file:
            if export_format == 'yaml':
                import yaml
                yaml.safe_dump(self.to_dict(), export_file, default_flow_style=False)
            elif export_format == 'json':
                export_file.write(text_type(json.dumps(self.to_dict(), ensure_ascii=False)))
            return True


def generate_add_user_command(proposed_user=None, manage_home=None):
    """Generate command to add a user.

    args:
        proposed_user (User): User
        manage_home: bool

    returns:
        list: The command string split into shell-like syntax
    """
    command = None
    if get_platform() in ('Linux', 'OpenBSD'):
        command = '{0} {1}'.format(sudo_check(), LINUX_CMD_USERADD)
        if proposed_user.uid:
            command = '{0} -u {1}'.format(command, proposed_user.uid)
        if proposed_user.gid:
            command = '{0} -g {1}'.format(command, proposed_user.gid)
        if proposed_user.gecos:
            command = '{0} -c \'{1}\''.format(command, proposed_user.gecos)
        if manage_home:
            if proposed_user.home_dir:
                if os.path.exists(proposed_user.home_dir):
                    command = '{0} -d {1}'.format(command, proposed_user.home_dir)
            elif not os.path.exists('/home/{0}'.format(proposed_user.name)):
                command = '{0} -m'.format(command)
        if proposed_user.shell:
            command = '{0} -s {1}'.format(command, proposed_user.shell)
        command = '{0} {1}'.format(command, proposed_user.name)
    elif get_platform() == 'FreeBSD':  # pragma: FreeBSD
        command = '{0} {1} useradd'.format(sudo_check(), FREEBSD_CMD_PW)
        if proposed_user.uid:
            command = '{0} -u {1}'.format(command, proposed_user.uid)
        if proposed_user.gid:
            command = '{0} -g {1}'.format(command, proposed_user.gid)
        if proposed_user.gecos:
            command = '{0} -c \'{1}\''.format(command, proposed_user.gecos)
        if manage_home:
            if proposed_user.home_dir:
                command = '{0} -d {1}'.format(command, proposed_user.home_dir)
            else:
                command = '{0} -m'.format(command)
        if proposed_user.shell:
            command = '{0} -s {1}'.format(command, proposed_user.shell)
        command = '{0} -n {1}'.format(command, proposed_user.name)

    if command:
        return shlex.split(str(command))


def generate_modify_user_command(task=None, manage_home=None):
    """Generate command to modify existing user to become the proposed user.

    args:
        task (dict): A proposed user and the differences between it and the existing user

    returns:
        list: The command string split into shell-like syntax
    """
    name = task['proposed_user'].name
    comparison_result = task['user_comparison']['result']
    command = None
    if get_platform() in ('Linux', 'OpenBSD'):
        command = '{0} {1}'.format(sudo_check(), LINUX_CMD_USERMOD)
        if comparison_result.get('replacement_uid_value'):
            command = '{0} -u {1}'.format(command, comparison_result.get('replacement_uid_value'))
        if comparison_result.get('replacement_gid_value'):
            command = '{0} -g {1}'.format(command, comparison_result.get('replacement_gid_value'))
        if comparison_result.get('replacement_gecos_value'):
            command = '{0} -c {1}'.format(command, comparison_result.get('replacement_gecos_value'))
        if comparison_result.get('replacement_shell_value'):
            command = '{0} -s {1}'.format(command, comparison_result.get('replacement_shell_value'))
        if manage_home and comparison_result.get('replacement_home_dir_value'):
                command = '{0} -d {1}'.format(command, comparison_result.get('replacement_home_dir_value'))
        command = '{0} {1}'.format(command, name)
    if get_platform() == 'FreeBSD':  # pragma: FreeBSD
        command = '{0} {1} usermod'.format(sudo_check(), FREEBSD_CMD_PW)
        if comparison_result.get('replacement_uid_value'):
            command = '{0} -u {1}'.format(command, comparison_result.get('replacement_uid_value'))
        if comparison_result.get('replacement_gid_value'):
            command = '{0} -g {1}'.format(command, comparison_result.get('replacement_gid_value'))
        if comparison_result.get('replacement_gecos_value'):
            command = '{0} -c {1}'.format(command, comparison_result.get('replacement_gecos_value'))
        if comparison_result.get('replacement_shell_value'):
            command = '{0} -s {1}'.format(command, comparison_result.get('replacement_shell_value'))
        if manage_home and comparison_result.get('replacement_home_dir_value'):
            command = '{0} -d {1}'.format(command, comparison_result.get('replacement_home_dir_value'))
        command = '{0} -n {1}'.format(command, name)
    if command:
        return shlex.split(str(command))


def generate_delete_user_command(username=None, manage_home=None):
    """Generate command to delete a user.

    args:
        username (str): user name
        manage_home (bool): manage home directory

    returns:
        list: The user delete command string split into shell-like syntax
    """
    command = None
    remove_home = '-r' if manage_home else ''

    if get_platform() in ('Linux', 'OpenBSD'):
        command = '{0} {1} {2} {3}'.format(sudo_check(), LINUX_CMD_USERDEL, remove_home, username)
    elif get_platform() == 'FreeBSD':  # pragma: FreeBSD
        command = '{0} {1} userdel {2} -n {3}'.format(sudo_check(), FREEBSD_CMD_PW, remove_home, username)
    if command:
        return shlex.split(str(command))


def get_user_by_uid(uid=None, users=None):
    """Return a list of users, from a supplied list, based on their uid.

    args:
        uid (id): A user id
        user_list (list): An instance of Users

    returns:
        list: a list of users matching the supplied uid
    """
    return users.describe_users(users_filter=dict(uid=uid))


def compare_user(passed_user=None, user_list=None):
    """Check if supplied User instance exists in supplied Users list and, if so, return the differences.

    args:
        passed_user (User): the user instance to check for differences
        user_list (Users): the Users instance containing a list of Users instances

    returns:
        dict: Details of the matching user and a list of differences
    """
    # Check if user exists
    returned = user_list.describe_users(users_filter=dict(name=passed_user.name))
    replace_keys = False
    # User exists, so compare attributes
    comparison_result = dict()
    if passed_user.uid and (not returned[0].uid == passed_user.uid):
        comparison_result['uid_action'] = 'modify'
        comparison_result['current_uid_value'] = returned[0].uid
        comparison_result['replacement_uid_value'] = passed_user.uid
    if passed_user.gid and (not returned[0].gid == passed_user.gid):
        comparison_result['gid_action'] = 'modify'
        comparison_result['current_gid_value'] = returned[0].gid
        comparison_result['replacement_gid_value'] = passed_user.gid
    if passed_user.gecos and (not returned[0].gecos == passed_user.gecos):
        comparison_result['gecos_action'] = 'modify'
        comparison_result['current_gecos_value'] = returned[0].gecos
        comparison_result['replacement_gecos_value'] = passed_user.gecos
    if passed_user.home_dir and (not returned[0].home_dir == passed_user.home_dir):
        comparison_result['home_dir_action'] = 'modify'
        comparison_result['current_home_dir_value'] = returned[0].home_dir
        comparison_result['replacement_home_dir_value'] = passed_user.home_dir
        # (Re)set keys if home dir changed
        replace_keys = True
    if passed_user.shell and (not returned[0].shell == passed_user.shell):
        comparison_result['shell_action'] = 'modify'
        comparison_result['current_shell_value'] = returned[0].shell
        comparison_result['replacement_shell_value'] = passed_user.shell
    if passed_user.sudoers_entry and (not returned[0].sudoers_entry == passed_user.sudoers_entry):
        comparison_result['sudoers_entry_action'] = 'modify'
        comparison_result['current_sudoers_entry'] = returned[0].sudoers_entry
        comparison_result['replacement_sudoers_entry'] = passed_user.sudoers_entry
    # if passed_user.public_keys and (not returned[0].public_keys == passed_user.public_keys):
    existing_keys = returned[0].public_keys
    passed_keys = passed_user.public_keys
    # Check if existing and passed keys exist, and if so, compare
    if all((existing_keys, passed_keys)) and len(existing_keys) == len(passed_user.public_keys):
        # Compare each key, and if any differences, replace
        existing = set(key.raw for key in existing_keys)
        replacement = set(key.raw for key in passed_keys)
        if set.difference(existing, replacement):
            replace_keys = True
    # If not existing keys but keys passed set, then
    elif passed_keys and not existing_keys:
        replace_keys = True
    if replace_keys:
        comparison_result['public_keys_action'] = 'modify'
        comparison_result['current_public_keys_value'] = existing_keys
        comparison_result['replacement_public_keys_value'] = passed_keys
    return dict(state='existing', result=comparison_result, existing_user=returned)



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, unicode_literals, print_function)

import pytest
from creds.ssh import PublicKey

from .sample_data import PUBLIC_KEYS


def test_invalid_public_key_setup():
    with pytest.raises(AttributeError):
        assert PublicKey()


def test_create_public_key_from_encoded():
    for key in PUBLIC_KEYS:
        public_key = PublicKey(
            b64encoded=key['encoded'])
        assert public_key.raw == key['raw']
        assert public_key.b64encoded == key['encoded']


def test_create_public_key_from_raw():
    for key in PUBLIC_KEYS:
        public_key = PublicKey(raw=key['raw'])
        assert public_key.b64encoded == key['encoded']
        assert public_key.raw == key['raw']


def test_public_key_repr_and_str():
    public_key = PublicKey(raw=PUBLIC_KEYS[0]['raw'])
    assert str(public_key) == public_key.__repr__()



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, unicode_literals, print_function)

import os

import pytest

from creds.ssh import PublicKey
from creds.users import (Users, User)
from creds.utils import sudo_check
from tests.sample_data import PUBLIC_KEYS
from .sample_data import SAMPLE_DICT


def test_users_yaml_export(tmpdir):
    """ Test the exporting of a Users sequence to yaml. """
    export_file = tmpdir.mkdir("export").join("export.yml")
    users = Users.from_dict(SAMPLE_DICT)
    assert users.export(file_path=export_file.strpath, export_format='yaml')
    exported_users = Users.from_yaml(export_file.strpath)
    for index, _ in enumerate(users):
        assert users[index].name == exported_users[index].name
        assert users[index].passwd == exported_users[index].passwd
        assert users[index].uid == exported_users[index].uid
        assert users[index].gid == exported_users[index].gid
        assert users[index].gecos == exported_users[index].gecos
        assert users[index].home_dir == exported_users[index].home_dir
        assert users[index].shell == exported_users[index].shell
        for pk_index, _ in enumerate(users[index].public_keys):
            assert users[index].public_keys[pk_index].raw == exported_users[index].public_keys[pk_index].raw
            assert users[index].public_keys[pk_index].b64encoded == exported_users[index].public_keys[
                pk_index].b64encoded


def test_users_json_export(tmpdir):
    """ Test the exporting of a Users sequence to yaml. """
    export_file = tmpdir.mkdir("export").join("export.json")
    users = Users.from_dict(SAMPLE_DICT)
    assert users.export(file_path=export_file.strpath, export_format='json')
    exported_users = Users.from_json(export_file.strpath)
    for index, _ in enumerate(users):
        assert users[index].name == exported_users[index].name
        assert users[index].passwd == exported_users[index].passwd
        assert users[index].uid == exported_users[index].uid
        assert users[index].gid == exported_users[index].gid
        assert users[index].gecos == exported_users[index].gecos
        assert users[index].home_dir == exported_users[index].home_dir
        assert users[index].shell == exported_users[index].shell
        for pk_index, _ in enumerate(users[index].public_keys):
            assert users[index].public_keys[pk_index].raw == exported_users[index].public_keys[pk_index].raw
            assert users[index].public_keys[pk_index].b64encoded == exported_users[index].public_keys[
                pk_index].b64encoded


def test_users_instance_creation():
    """ Test creation of instances of User and add to Users collection. """
    input_user_list = Users()
    input_user_list.append(
        User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh'))
    input_user_list.append(
        User(name='jane', uid=1002, gid=1002, gecos='jane comment', home_dir='/home/jane', shell='/bin/bash'))
    input_user_list.append(
        User(name='freddy', uid=1003, gid=1003, gecos='freddy comment', home_dir='/home/freddy', shell='/bin/false'))
    assert len(input_user_list) == 3


def test_users_del_method():
    users = Users()
    users.append(
        User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh'))
    users.append(
        User(name='jane', uid=1002, gid=1002, gecos='jane comment', home_dir='/home/jane', shell='/bin/sh'))
    assert len(users) == 2
    del users[0]
    assert len(users) == 1


def test_users_insert_method():
    users = Users()
    users.append(
        User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh'))
    users.append(
        User(name='jane', uid=1002, gid=1002, gecos='jane comment', home_dir='/home/jane', shell='/bin/sh'))
    users.insert(0, User(name='freddy', uid=1003, gid=1003, gecos='freddy comment',
                         home_dir='/home/freddy', shell='/bin/false'))
    assert len(users) == 3
    with pytest.raises(TypeError):
        users.insert(0, dict(name='freddy', uid=1003, gid=1003, gecos='freddy comment',
                             home_dir='/home/freddy', shell='/bin/false'))


def test_get_users_from_passwd():
    """ Test creation of a Users collection based on users found in the passwd file. """
    users = Users.from_passwd()
    assert isinstance(users, Users)


def test_get_users_from_dict():
    """ Test creation of a Users collection based on a predefined dict. """
    users = Users.from_dict(input_dict=SAMPLE_DICT)
    assert isinstance(users, Users)
    assert isinstance(users[0], User)
    assert isinstance(users[0].uid, int)


def test_get_users_from_yaml():
    """ Test creation of a Users collection based on a yaml document. """
    users = Users.from_yaml(file_path='{0}/yaml_input/basic.yml'.format(os.path.dirname(os.path.abspath(__file__))))
    assert isinstance(users, Users)
    assert isinstance(users[0], User)
    assert isinstance(users[0].uid, int)
    assert users[0].name == 'peter'
    assert users[0].home_dir == '/home/bigal'


def test_get_users_from_json():
    """ Test creation of a Users collection based on a json document. """
    users = Users.from_json(file_path='{0}/json_input/basic.json'.format(os.path.dirname(os.path.abspath(__file__))))
    assert isinstance(users, Users)
    assert isinstance(users[0], User)
    assert isinstance(users[0].uid, int)


def test_get_users_from_invalid_yaml():
    """ Test a ValueError is raised if loading a yaml file of users with invalid syntax. """
    with pytest.raises(ValueError):
        Users.from_yaml(file_path='{0}/yaml_input/invalid.yml'.format(os.path.dirname(os.path.abspath(__file__))))


def test_get_users_from_invalid_json():
    """ Test a ValueError is raised if loading a json file of users with invalid syntax. """
    with pytest.raises(ValueError):
        Users.from_json(file_path='{0}/json_input/invalid.json'.format(os.path.dirname(os.path.abspath(__file__))))


def test_users_repr():
    users = Users()
    users.append(User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh'))
    assert str(users) == users.__repr__()


def test_users_add_and_remove():
    rod = User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh')
    users = Users()
    users.append(rod)
    assert len(users) == 1
    jane = User(name='jane')
    users.append(jane)
    assert len(users) == 2
    users.remove(username='jane')
    assert len(users) == 1


def test_users_set_item():
    rod = User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh')
    users = Users()
    users.append(rod)
    users[0] = User(name='jane', uid=1002, gid=1002, gecos='jane comment', home_dir='/home/jane', shell='/bin/sh')
    assert len(users) == 1
    assert users[0].name == 'jane'


def test_users_filters():
    users = Users()
    users.append(User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh'))
    assert not users.describe_users(users_filter=dict(name='nobody'))
    assert not users.describe_users(users_filter=dict(uid=1000))


def test_user_instance_creation():
    name = 'Fred'
    uid = 1024
    gid = 1024
    gecos = 'Fred Bloggs'
    home_dir = '/home/fred'
    shell = '/bin/false'
    public_key = PublicKey(raw=PUBLIC_KEYS[0]['raw'])
    test_user = User(name=name, uid=uid, gid=gid, gecos=gecos, home_dir=home_dir, shell=shell, public_keys=[public_key])
    assert test_user.name == name
    assert test_user.uid == uid
    assert test_user.gid == gid
    # Ensure gecos is surrounded with double quotes
    assert test_user.gecos.startswith('\"') and test_user.gecos.endswith('\"')
    assert test_user.home_dir == home_dir
    assert test_user.shell == shell
    assert test_user.public_keys == [public_key]


def test_user_instance_creation_precommented_gecos():
    name = 'Fred'
    uid = 1024
    gid = 1024
    gecos = '\'Fred Bloggs\''
    home_dir = '/home/fred'
    shell = '/bin/false'
    test_user = User(name=name, uid=uid, gid=gid, gecos=gecos, home_dir=home_dir, shell=shell)
    assert test_user.name == name
    assert test_user.uid == uid
    assert test_user.gid == gid
    # Ensure gecos is surrounded with double quotes
    assert test_user.gecos.startswith('\"') and test_user.gecos.endswith('\"')
    assert test_user.home_dir == home_dir
    assert test_user.shell == shell


def test_user_instance_with_missing_gecos():
    rod = User(name='rod', uid=1001, gid=1001, home_dir='/home/rod', shell='/bin/sh')
    assert rod.gecos == None


def test_platform_detection(monkeypatch):
    monkeypatch.setattr("platform.system", lambda: 'Darwin')
    with pytest.raises(SystemExit):
        Users()


def test_user_detection(monkeypatch):
    monkeypatch.setattr("os.geteuid", lambda: 1)
    assert sudo_check().endswith('sudo')
    monkeypatch.setattr("os.geteuid", lambda: 0)
    assert sudo_check() == ''



# -*- coding: utf-8 -*-

from __future__ import (absolute_import, unicode_literals)

import getpass
import os
import shlex

import pytest
# import yaml
# from boto3 import Session
# from moto import mock_s3

from creds.constants import (LINUX_CMD_USERADD, LINUX_CMD_USERDEL,
                             LINUX_CMD_GROUP_ADD, FREEBSD_CMD_PW)
from creds.plan import (create_plan, execute_plan)
from creds.ssh import PublicKey
from creds.users import (Users, User)
from creds.utils import (execute_command, sudo_check, get_platform, remove_sudoers_entry)
from external.six import text_type
from .sample_data import PUBLIC_KEYS

# TODO: Detect based on OS
USERMOD = '/usr/sbin/usermod'
USERADD = '/usr/sbin/useradd'
USERDEL = '/usr/sbin/userdel'
GROUPADD = '/usr/sbin/groupadd'
GROUPDEL = '/usr/sbin/groupdel'

PLATFORM = get_platform()

CURRENT_USER = getpass.getuser()


# @mock_s3
# def test_execute_plan_to_create_user_with_downloaded_yaml():
#     """ Create a new user from downloaded YAML file """
#     delete_test_user_and_group()
#     session = Session()
#     s3_client = session.client('s3')
#     test_user_1 = open(os.path.join(os.path.dirname(__file__), 'test_user_1.yml')).read()
#     s3_client.create_bucket(Bucket='test')
#     s3_client.put_object(Bucket='test', Key='test.yml', Body=test_user_1)
#     response = s3_client.get_object(Bucket='test', Key='test.yml')
#     contents = response['Body'].read()
#     yaml_content = yaml.load(contents)
#     current_users = Users.from_passwd()
#     provided_users = Users.from_dict(yaml_content)
#     plan = create_plan(existing_users=current_users, proposed_users=provided_users, manage_home=False,
#                        protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
#     execute_plan(plan=plan)
#     updated_users = Users.from_passwd()
#     updated_user = updated_users.describe_users(users_filter=dict(name='dummy'))
#     assert len(updated_user) == 1
#     assert updated_user[0].name == 'dummy'
#     assert updated_user[0].gecos == '\"dummy (user) test\"'
#     assert not updated_user[0].public_keys
#     assert updated_user[0].sudoers_entry == 'ALL=(ALL:ALL) ALL'
#     delete_test_user_and_group()


def test_execute_plan_to_create_user_with_invalid_sudoers_entry():
    """ Create a new user but specify an invalid sudoers entry """
    delete_test_user_and_group()
    raw_public_key_2 = PUBLIC_KEYS[1].get('raw')
    public_key_2 = PublicKey(raw=raw_public_key_2)
    current_users = Users.from_passwd()
    provided_users = Users()
    provided_users.append(
        User(name='testuserx1234', uid=59998, gid=1, gecos='test user gecos update',
             shell='/bin/false', public_keys=[public_key_2], sudoers_entry='INVALID ALL=(ALL:ALL) ALL'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, manage_home=False,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert plan[0]['proposed_user'].gecos == '\"test user gecos update\"'
    with pytest.raises(ValueError):
        execute_plan(plan=plan)
    delete_test_user_and_group()


def test_execute_plan_to_delete_user_ignoring_home():
    """ Delete a user and ensure their home dir is untouched """

    delete_test_user_and_group()
    pre_users = Users.from_passwd()
    create_test_user()
    plan = create_plan(existing_users=Users.from_passwd(), proposed_users=pre_users, purge_undefined=True,
                       manage_home=False,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant',
                                        CURRENT_USER])
    execute_plan(plan=plan)
    updated_users = Users.from_passwd()
    updated_user = updated_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert len(updated_user) == 0
    assert os.path.exists('/home/testuserx1234')
    delete_test_user_and_group()


def test_execute_plan_to_create_user_ignoring_home():
    """ Create a new user without creating home directory """

    delete_test_user_and_group()
    raw_public_key_2 = PUBLIC_KEYS[1].get('raw')
    public_key_2 = PublicKey(raw=raw_public_key_2)
    current_users = Users.from_passwd()
    provided_users = Users()
    provided_users.append(
        User(name='testuserx1234', uid=59998, gid=1, gecos='test user gecos update',
             shell='/bin/false', public_keys=[public_key_2], sudoers_entry='ALL=(ALL:ALL) ALL'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, manage_home=False,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert plan[0]['proposed_user'].gecos == '\"test user gecos update\"'
    execute_plan(plan=plan)
    updated_users = Users.from_passwd()
    updated_user = updated_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert len(updated_user) == 1
    assert updated_user[0].name == 'testuserx1234'
    assert updated_user[0].uid == 59998
    assert updated_user[0].gid == 1
    assert updated_user[0].gecos == '\"test user gecos update\"'
    assert updated_user[0].shell == '/bin/false'
    assert not updated_user[0].public_keys
    assert updated_user[0].sudoers_entry == 'ALL=(ALL:ALL) ALL'
    delete_test_user_and_group()


def test_execute_plan_to_update_existing_user_ignoring_keys():
    """ Create a new user without touching keys """

    delete_test_user_and_group()
    create_test_user()
    raw_public_key_2 = PUBLIC_KEYS[1].get('raw')
    public_key_2 = PublicKey(raw=raw_public_key_2)
    current_users = Users.from_passwd()
    provided_users = Users()
    provided_users.append(
        User(name='testuserx1234', uid=59998, gid=1, gecos='test user gecos update',
             shell='/bin/false', public_keys=[public_key_2], sudoers_entry='ALL=(ALL:ALL) ALL'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, manage_keys=False,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert plan[0]['proposed_user'].gecos == '\"test user gecos update\"'
    execute_plan(plan=plan)
    updated_users = Users.from_passwd()
    updated_user = updated_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert len(updated_user) == 1
    assert updated_user[0].name == 'testuserx1234'
    assert updated_user[0].uid == 59998
    assert updated_user[0].gid == 1
    assert updated_user[0].gecos == '\"test user gecos update\"'
    assert updated_user[0].shell == '/bin/false'
    assert not updated_user[0].public_keys
    assert updated_user[0].sudoers_entry == 'ALL=(ALL:ALL) ALL'
    delete_test_user_and_group()


def test_execute_plan_to_update_existing_user():
    """ Create a new user and then attempt to create another user with existing id """

    delete_test_user_and_group()
    create_test_user()
    raw_public_key_2 = PUBLIC_KEYS[1].get('raw')
    public_key_2 = PublicKey(raw=raw_public_key_2)
    current_users = Users.from_passwd()
    provided_users = Users()
    provided_users.append(
        User(name='testuserx1234', uid=59998, gid=1, gecos='test user gecos update',
             shell='/bin/false', public_keys=[public_key_2], sudoers_entry='ALL=(ALL:ALL) ALL'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert plan[0]['proposed_user'].gecos == '\"test user gecos update\"'
    execute_plan(plan=plan)
    updated_users = Users.from_passwd()
    updated_user = updated_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert len(updated_user) == 1
    assert updated_user[0].name == 'testuserx1234'
    assert updated_user[0].uid == 59998
    assert updated_user[0].gid == 1
    assert updated_user[0].gecos == '\"test user gecos update\"'
    assert updated_user[0].shell == '/bin/false'
    assert updated_user[0].public_keys[0].raw == text_type(PUBLIC_KEYS[1]['raw'])
    assert updated_user[0].sudoers_entry == 'ALL=(ALL:ALL) ALL'
    delete_test_user_and_group()


def test_create_and_execute_plan_to_create_new_user_with_sudo_all():
    """ Test creation of a user instance with sudo all and then write """
    delete_test_user_and_group()
    create_test_group()
    current_users = Users.from_passwd()
    provided_users = Users()

    public_keys = [PublicKey(
        b64encoded=PUBLIC_KEYS[0]['encoded'])]
    provided_users.append(
        User(name='testuserx1234', home_dir='/home/testuserx1234', shell='/bin/false', gid=59999, uid=59999,
             gecos='test user gecos',
             public_keys=public_keys, sudoers_entry='ALL=(ALL)\tNOPASSWD:ALL'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, purge_undefined=True,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'vagrant', CURRENT_USER])
    assert plan[0]['state'] == 'missing'
    assert plan[0]['proposed_user'].name == "testuserx1234"
    assert plan[0]['proposed_user'].home_dir == "/home/testuserx1234"
    assert plan[0]['proposed_user'].uid == 59999
    assert plan[0]['proposed_user'].gid == 59999
    assert plan[0]['proposed_user'].gecos == '\"test user gecos\"'
    assert plan[0]['proposed_user'].shell == '/bin/false'
    assert plan[0]['proposed_user'].sudoers_entry == 'ALL=(ALL)\tNOPASSWD:ALL'
    assert type(plan[0]['proposed_user'].public_keys[0].raw) == text_type
    assert plan[0]['proposed_user'].public_keys[0].raw == text_type(PUBLIC_KEYS[0]['raw'])
    execute_plan(plan=plan)
    current_users = Users.from_passwd()
    created_user = current_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert created_user[0].sudoers_entry == 'ALL=(ALL)\tNOPASSWD:ALL'
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, purge_undefined=True,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'vagrant', CURRENT_USER])
    assert not plan


def test_users_instance_creation():
    """ Test creation of a user instance """
    users = Users()
    users.append(
        User(name='rod', uid=1001, gid=1001, gecos='rod comment', home_dir='/home/rod', shell='/bin/sh'))
    users.append(
        User(name='jane', uid=1002, gid=1002, gecos='jane comment', home_dir='/home/jane', shell='/bin/bash'))
    users.append(
        User(name='freddy', uid=1003, gid=1003, gecos='freddy comment', home_dir='/home/freddy', shell='/bin/false'))
    assert len(users) == 3


def test_create_and_execute_plan_to_create_new_user():
    """ Test creation of user instance and then write """
    delete_test_user_and_group()
    create_test_group()
    current_users = Users.from_passwd()

    provided_users = Users()

    public_keys = [PublicKey(
        b64encoded=PUBLIC_KEYS[0]['encoded'])]
    provided_users.append(
        User(name='testuserx1234', home_dir='/home/testuserx1234', shell='/bin/false', gid=59999, uid=59999,
             gecos='test user gecos',
             public_keys=public_keys))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, purge_undefined=True,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'vagrant', CURRENT_USER])
    assert plan[0]['state'] == 'missing'
    assert plan[0]['proposed_user'].name == "testuserx1234"
    assert plan[0]['proposed_user'].home_dir == "/home/testuserx1234"
    assert plan[0]['proposed_user'].uid == 59999
    assert plan[0]['proposed_user'].gid == 59999
    assert plan[0]['proposed_user'].gecos == '\"test user gecos\"'
    assert plan[0]['proposed_user'].shell == '/bin/false'
    assert type(plan[0]['proposed_user'].public_keys[0].raw) == text_type
    assert plan[0]['proposed_user'].public_keys[0].raw == text_type(PUBLIC_KEYS[0]['raw'])
    execute_plan(plan=plan)

    current_users = Users.from_passwd()
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, purge_undefined=True,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert not plan


def test_create_and_execute_plan_to_create_identical_user():
    delete_test_user_and_group()
    create_test_user()
    current_users = Users.from_passwd()
    provided_users = Users()
    provided_users.append(User(name='testuserx1234', uid=59999, gecos='test user gecos'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    execute_plan(plan=plan)
    current_users = Users.from_passwd()
    plan = create_plan(existing_users=current_users, proposed_users=provided_users,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert not plan
    delete_test_user_and_group()


def test_update_existing_user():
    delete_test_user_and_group()
    create_test_user()
    current_users = Users.from_passwd()
    provided_users = Users()
    raw_public_key = PUBLIC_KEYS[0].get('raw')
    public_key = PublicKey(raw=raw_public_key)
    provided_users.append(
        User(name='testuserx1234', uid=59999, gecos='test user gecos update', home_dir='/tmp/temp',
             public_keys=[public_key]))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert plan[0]['action'] == 'update'
    execute_plan(plan)
    current_users = Users.from_passwd()
    new_user = current_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert new_user[0].public_keys[0].raw == raw_public_key
    delete_test_user_and_group()


def test_execute_plan_to_create_new_user_with_clashing_uid():
    """ Create a new user and then attempt to create another user with existing id """
    delete_test_user_and_group()
    current_users = Users.from_passwd()
    provided_users = Users()
    provided_users.append(User(name='testuserx1234', uid=59999, gecos='test user gecos'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert plan[0]['action'] == 'add'
    assert plan[0]['proposed_user'].name == "testuserx1234"
    assert plan[0]['proposed_user'].uid == 59999
    assert plan[0]['proposed_user'].gecos == '\"test user gecos\"'
    execute_plan(plan=plan)
    current_users = Users.from_passwd()
    provided_users = Users()
    provided_users.append(User(name='testuserx12345', uid=59999, gecos='test user gecos'))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users, purge_undefined=True,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    assert plan[0]['error'] == 'uid_clash'
    execute_plan(plan=plan)
    delete_test_user_and_group()


def test_execute_plan_to_update_existing_user_with_multiple_keys():
    """ Create a new user with 2 keys and then replace with a new one """
    create_test_user()
    raw_public_key_1 = PUBLIC_KEYS[0].get('raw')
    public_key_1 = PublicKey(raw=raw_public_key_1)
    raw_public_key_2 = PUBLIC_KEYS[1].get('raw')
    public_key_2 = PublicKey(raw=raw_public_key_2)
    raw_public_key_3 = PUBLIC_KEYS[2].get('raw')
    public_key_3 = PublicKey(raw=raw_public_key_3)
    raw_public_key_4 = PUBLIC_KEYS[3].get('raw')
    public_key_4 = PublicKey(raw=raw_public_key_4)
    current_users = Users.from_passwd()
    provided_users_2 = Users()
    provided_users_2.append(User(name='testuserx1234', uid=59998, gid=1, gecos='test user gecos update',
                                 shell='/bin/false', public_keys=[public_key_1, public_key_2]))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users_2,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    execute_plan(plan=plan)
    updated_users = Users.from_passwd()
    updated_user = updated_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert updated_user[0].public_keys[0].raw == text_type(PUBLIC_KEYS[0]['raw'])
    assert updated_user[0].public_keys[1].raw == text_type(PUBLIC_KEYS[1]['raw'])
    # Replace both keys
    current_users = Users.from_passwd()
    provided_users_3 = Users()
    provided_users_3.append(User(name='testuserx1234', uid=59998, gid=1, gecos='test user gecos update',
                                 shell='/bin/false', public_keys=[public_key_3, public_key_4]))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users_3,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    execute_plan(plan=plan)
    updated_users = Users.from_passwd()
    updated_user = updated_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert updated_user[0].public_keys[0].raw == text_type(PUBLIC_KEYS[2]['raw'])
    assert updated_user[0].public_keys[1].raw == text_type(PUBLIC_KEYS[3]['raw'])
    # Replace one key
    current_users = Users.from_passwd()
    provided_users_4 = Users()
    provided_users_4.append(
        User(name='testuserx1234', uid=59998, gid=1, gecos='test user gecos update',
             shell='/bin/false', public_keys=[public_key_2, public_key_4]))
    plan = create_plan(existing_users=current_users, proposed_users=provided_users_4,
                       protected_users=['travis', 'couchdb', 'ubuntu', 'nginx', 'hadfielj', 'vagrant', CURRENT_USER])
    execute_plan(plan=plan)
    updated_users = Users.from_passwd()
    updated_user = updated_users.describe_users(users_filter=dict(name='testuserx1234'))
    assert updated_user[0].public_keys[0].raw == text_type(PUBLIC_KEYS[1]['raw'])
    assert updated_user[0].public_keys[1].raw == text_type(PUBLIC_KEYS[3]['raw'])
    delete_test_user_and_group()


def delete_test_user_and_group():
    if PLATFORM == 'Linux':
        del_user_command = shlex.split(str('{0} {1} -r -f testuserx1234'.format(sudo_check(), LINUX_CMD_USERDEL)))
        execute_command(command=del_user_command)
    elif PLATFORM == 'OpenBSD':
        del_user_command = shlex.split(str('{0} {1} -r testuserx1234'.format(sudo_check(), LINUX_CMD_USERDEL)))
        execute_command(command=del_user_command)
    elif PLATFORM == 'FreeBSD':
        del_user_command = shlex.split(str('{0} {1} userdel -r -n testuserx1234'.format(sudo_check(), FREEBSD_CMD_PW)))
        execute_command(command=del_user_command)
    if PLATFORM in ('Linux', 'OpenBSD'):
        del_group_command = shlex.split(str('{0} {1} testuserx1234'.format(sudo_check(), GROUPDEL)))
        execute_command(command=del_group_command)
        del_user_ssh_dir_command = shlex.split(str('/bin/rm -rf /tmp/.ssh'))
        execute_command(command=del_user_ssh_dir_command)
    remove_sudoers_entry(username='testuserx1234')
    execute_command(command=shlex.split(str('{0} rm -rf /home/testuserx1234'.format(sudo_check()))))


def create_test_user():
    if PLATFORM in ('Linux', 'OpenBSD'):
        command = shlex.split(
            str('{0} {1} -u 59999 -c \"test user gecos\" -m  -s /bin/bash testuserx1234'.format(sudo_check(),
                                                                                                LINUX_CMD_USERADD)))
    elif PLATFORM == 'FreeBSD':
        command = shlex.split(
            str('{0} {1} useradd -u 59999 -c \"test user gecos\" -m  -s /bin/bash -n testuserx1234'.format(sudo_check(),
                                                                                                           FREEBSD_CMD_PW)))
    assert execute_command(command=command)


def create_test_group():
    if PLATFORM in ('Linux', 'OpenBSD'):
        command = shlex.split(
            str('{0} {1} -g 59999 testuserx1234'.format(sudo_check(), LINUX_CMD_GROUP_ADD)))
    elif PLATFORM == 'FreeBSD':
        command = shlex.split(
            str('{0} {1} groupadd -g 59999 -n testuserx1234'.format(sudo_check(), FREEBSD_CMD_PW)))
    assert execute_command(command=command)



#!/usr/bin/env python

import os
import re
import sys
# from codecs import open

from setuptools import (setup, find_packages)
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass into py.test")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        import pytest

        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload -r pypi')
    sys.exit()

requires = []
test_requirements = ['pytest>=2.9.2', 'pytest-cov>=2.3.1', 'PyYAML>=3.11', 'boto', 'boto3', 'moto', 'mock']

with open('lib/creds/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

readme = open('README.rst').read()
long_description = readme

setup(
    name='creds',
    version=version,
    description='Creds is a library for managing Linux, FreeBSD and OpenBSD user accounts and credentials.',
    long_description=long_description,
    author='Jon Hadfield',
    author_email='jon@lessknown.co.uk',
    url='http://github.com/jonhadfield/creds',
    packages=find_packages('lib'),
    package_dir={'': 'lib'},
    # package_data={'': ['LICENSE', 'NOTICE'], 'creds': ['*.pem']},
    include_package_data=True,
    install_requires=requires,
    license='MIT',
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: System Administrators',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: BSD :: FreeBSD',
        'Operating System :: POSIX :: BSD :: OpenBSD',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy'
    ],
    cmdclass={'test': PyTest},
    tests_require=test_requirements,
    # extras_require={
    #     'security': [],
    # },
)



# -*- coding: utf-8 -*-
from typing import Tuple, List

UID_MIN: int
UID_MAX: int
SUPPORTED_PLATFORMS: List[str]
RANDOM_FILE_EXT_LENGTH: int
PURGE_UNDEFINED: bool
DEFAULT_UID_MIN: int
DEFAULT_UID_MAX: int
CMD_SUDO: str
LINUX_CMD_USERADD: str
LINUX_CMD_USERMOD: str
LINUX_CMD_USERDEL: str
LINUX_CMD_GROUP_ADD: str
LINUX_CMD_GROUP_DEL: str
LINUX_CMD_VISUDO: str
FREEBSD_CMD_PW: str


def login_defs() -> Tuple: pass
